@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1998-2000 Apple Computer, Inc. All rights reserved.
+ * Copyright (c) 1998-2002 Apple Computer, Inc. All rights reserved.
  *
  * @APPLE_LICENSE_HEADER_START@
  * 
@@ -61,14 +61,18 @@ OSDefineMetaClassAndStructors(MacRISC2PE, ApplePlatformExpert);
 
 bool MacRISC2PE::start(IOService *provider)
 {
-    long            	machineType;
-    OSData          	*tmpData;
-    IORegistryEntry 	*uniNRegEntry;
-    IORegistryEntry 	*powerMgtEntry;
-    UInt32			   	*primInfo;
-    UInt32			   	uniNArbCtrl, uniNBaseAddressTemp;
-	UInt32				stepType;
-    const char 			*provider_name;
+    long            		machineType;
+    OSData          		*tmpData;
+    IORegistryEntry 		*uniNRegEntry;
+    IORegistryEntry 		*powerMgtEntry;
+    UInt32			   		*primInfo;
+    UInt32			   		uniNArbCtrl, uniNBaseAddressTemp, uniNMPCIMemTimeout;
+	UInt32					stepType;
+	OSDictionary			*propTable;
+	OSCollectionIterator	*propIter;
+	OSString				*propKey;
+	OSData					*propVal;
+	SInt32					numFuncs, retval;
 	
     setChipSetType(kChipSetTypeCore2001);
 	
@@ -81,6 +85,8 @@ bool MacRISC2PE::start(IOService *provider)
 	if (provider_name != NULL) {
 		if (0 == strncmp(provider_name, "PowerMac", strlen("PowerMac")))
 			machineType = kMacRISC2TypePowerMac;
+		else if (0 == strncmp(provider_name, "RackMac", strlen("RackMac")))
+			machineType = kMacRISC2TypePowerMac;
 		else if (0 == strncmp(provider_name, "PowerBook", strlen("PowerBook")))
 			machineType = kMacRISC2TypePowerBook;
 		else if (0 == strncmp(provider_name, "iBook", strlen("iBook")))
@@ -104,7 +110,7 @@ bool MacRISC2PE::start(IOService *provider)
     tmpData = OSDynamicCast(OSData, uniNRegEntry->getProperty("reg"));
     if (tmpData == 0) return false;
     uniNBaseAddressTemp = ((unsigned long *)tmpData->getBytesNoCopy())[0];
-    uniNBaseAddress = (unsigned long *)ml_io_map(uniNBaseAddressTemp, 0x1000);
+    uniNBaseAddress = (unsigned long *)ml_io_map(uniNBaseAddressTemp, 0x3200);
     if (uniNBaseAddress == 0) return false;
   
     // Set QAckDelay depending on the version of Uni-N.
@@ -126,7 +132,48 @@ bool MacRISC2PE::start(IOService *provider)
         writeUniNReg(kUniNArbCtrl, uniNArbCtrl);
     }
 
-    // Creates the nubs for the children of uni-n
+    // Set Max/PCI Memory Timeout for appropriate Uni-N.   
+    if ( ((uniNVersion >= kUniNVersion150) && (uniNVersion <= kUniNVersion200)) || 
+         (uniNVersion == kUniNVersionPangea) )
+    {
+        uniNMPCIMemTimeout = readUniNReg(kUniNMPCIMemTimeout);
+        uniNMPCIMemTimeout &= ~kUniNMPCIMemTimeoutMask;
+        uniNMPCIMemTimeout |= kUniNMPCIMemGrantTime;
+        writeUniNReg(kUniNMPCIMemTimeout, uniNMPCIMemTimeout);
+    }
+	
+	// Parse platform function commands for uni-n
+	numFuncs = 0;
+	
+	if ( ((propTable = uniNRegEntry->dictionaryWithProperties()) == 0) ||
+	     ((propIter = OSCollectionIterator::withCollection(propTable)) == 0) ) {
+		if (propTable) propTable->release();
+		return(false);
+	}
+
+	while ((propKey = OSDynamicCast(OSString, propIter->getNextObject())) != 0) {
+		if (strncmp(kFunctionProvidedPrefix,
+		            propKey->getCStringNoCopy(),
+		            strlen(kFunctionProvidedPrefix)) == 0) {
+			propVal = OSDynamicCast(OSData, propTable->getObject(propKey));
+
+			if ((retval = parseProvidedFunction(propKey, propVal)) < 0) {
+				// something went wrong, free resources and bail out
+				releaseResources();
+				numFuncs = 0;
+				break;
+ 			} else
+ 				numFuncs += retval;
+ 		}
+    }
+	
+	if (propTable) { propTable->release(); propTable = 0; }
+	if (propIter) { propIter->release(); propIter = 0; }
+
+	// publish on-demand and services in IOResources
+	if (fOnDemand) publishStrings(fOnDemand);
+
+   // Creates the nubs for the children of uni-n
     IOService *uniNServiceEntry = OSDynamicCast(IOService, uniNRegEntry);
     if (uniNServiceEntry != NULL)
         createNubs(this, uniNRegEntry->getChildIterator( gIODTPlane ));
@@ -163,6 +210,7 @@ bool MacRISC2PE::start(IOService *provider)
 	
     // Set up processorSpeedChangeFlags depending on platform
 	processorSpeedChangeFlags = kNoSpeedChange;
+	stepType = 0;
     if (machineType == kMacRISC2TypePowerBook) {
 		OSIterator 		*childIterator;
 		IORegistryEntry *cpuEntry, *powerPCEntry;
@@ -192,8 +240,6 @@ bool MacRISC2PE::start(IOService *provider)
 
 							if (stepTypeData->getLength() > 0)
 								stepType = *(UInt32 *) stepTypeData->getBytesNoCopy();
-							else
-								stepType = 0;
 				
 							newCPUSpeed = *(UInt32 *) cpuSpeedData->getBytesNoCopy();
 							if (newCPUSpeed != gPEClockFrequencyInfo.cpu_clock_rate_hz) {
@@ -357,6 +403,77 @@ bool MacRISC2PE::platformAdjustService(IOService *service)
         return true;
     }
     
+	/*
+	 * For uni-n pci version 1.5 which services mac-io, tell pci driver to disable read data gating
+	 * -- Note that previous versions of this driver defined version 1.5 as 0x0010.  This is not 
+	 * correct and we now use 0x0011 for uni-n 1.5.
+	 */
+	if ((uniNVersion == kUniNVersion150) && IODTMatchNubWithKeys(service, "('pci', 'uni-north')") && 
+		(service->childFromPath("mac-io", gIODTPlane) != NULL)) {
+			service->setProperty ("DisableRDG", true);
+			return true;
+	}
+    
+	/* 
+	 * For Firewire on uni-n Pangea & uni-n 1.5 through 2.0, adjust the cache line size and timer latency
+	 */
+	if (((uniNVersion >= kUniNVersion150) && (uniNVersion <= kUniNVersion200) || (uniNVersion == kUniNVersionPangea)) &&
+		(!strcmp(service->getName(gIODTPlane), "firewire")) &&
+		IODTMatchNubWithKeys(service->getParentEntry(gIODTPlane), "('pci', 'uni-north')")) {
+			char data;
+			
+			data = 0x08;		// 32 byte cache line
+			service->setProperty (kIOPCICacheLineSize, &data, 1);
+			data = 0x40;		// latency timer to 40
+			service->setProperty (kIOPCITimerLatency, &data, 1);
+			return true;
+	}
+    
+	/* 
+	 * For Ethernet on uni-n 2.0, adjust the cache line size and timer latency
+	 */
+	if ((uniNVersion == kUniNVersion200) &&
+		IODTMatchNubWithKeys(service, "('ethernet', 'gmac')") &&
+		IODTMatchNubWithKeys(service->getParentEntry(gIODTPlane), "('pci', 'uni-north')")) {
+			char 	data;
+			long	cacheSize;
+			OSData 	*cacheData;
+			
+			data = 0x08; 	// assume default of 32 byte cache line
+			cacheData = OSDynamicCast( OSData, service->getProperty( "cache-line-size" ) );
+			if (cacheData) {
+				cacheSize = *(long *)cacheData->getBytesNoCopy();
+				data = (cacheSize >> 2);
+			}
+			service->setProperty (kIOPCICacheLineSize, &data, 1);
+			data = 0x20;		// latency timer to 20
+			service->setProperty (kIOPCITimerLatency, &data, 1);
+			return true;
+	}
+	
+	// For usb@18 on PowerBook4,x machines add an "AAPL,SuspendablePorts" property if it doesn't exist
+	if (!strcmp(service->getName(), "usb") && 
+		(0 == strncmp(provider_name, "PowerBook4,", strlen("PowerBook4,"))) &&
+		IODTMatchNubWithKeys(service->getParentEntry(gIODTPlane), "('pci', 'uni-north')")) {
+		
+		OSData				*regProp;
+		IOPCIAddressSpace	*pciAddress;
+		UInt32				ports;
+	
+		if( (regProp = (OSData *) service->getProperty("reg"))) {
+			pciAddress = (IOPCIAddressSpace *) regProp->getBytesNoCopy();
+			// Only for usb@18
+			if (pciAddress->s.deviceNum == 0x18) {
+				// If property doesn't exist, create it
+				if(!((OSData *) service->getProperty(kAAPLSuspendablePorts))) {
+					ports = 4;
+					service->setProperty (kAAPLSuspendablePorts, &ports, sizeof(UInt32));
+				}
+				return true;
+			}
+		}
+	}
+    
     return true;
 }
 
@@ -365,6 +482,8 @@ IOReturn MacRISC2PE::callPlatformFunction(const OSSymbol *functionName,
 					void *param1, void *param2,
 					void *param3, void *param4)
 {
+	OSData		*func;
+
     if (functionName == gGetDefaultBusSpeedsKey)
     {
         getDefaultBusSpeeds((long *)param1, (unsigned long **)param2);
@@ -373,12 +492,12 @@ IOReturn MacRISC2PE::callPlatformFunction(const OSSymbol *functionName,
   
     if (functionName->isEqualTo("EnableUniNEthernetClock"))
     {
-        enableUniNEthernetClock((bool)param1);
+        enableUniNEthernetClock((bool)param1, (IOService *)param2);
         return kIOReturnSuccess;
     }
 
     if (functionName->isEqualTo("EnableFireWireClock")) {
-        enableUniNFireWireClock((bool)param1);
+        enableUniNFireWireClock((bool)param1, (IOService *)param2);
         return kIOReturnSuccess;
     }
   
@@ -401,17 +520,25 @@ IOReturn MacRISC2PE::callPlatformFunction(const OSSymbol *functionName,
 		return platformPowerMonitor ((UInt32 *) param1);
     }
   
+	// Is this an on-demand platform function?
+	if (fOnDemand && (func = OSDynamicCast(OSData, fOnDemand->getObject(functionName))) != 0) {
+		if (performFunction(func, param1, param2, param3, param4))
+			return kIOReturnSuccess;
+		else
+			return kIOReturnError;
+	}
+
     return super::callPlatformFunction(functionName, waitForFunction, param1, param2, param3, param4);
 }
 
-unsigned long MacRISC2PE::readUniNReg(unsigned long offest)
+unsigned long MacRISC2PE::readUniNReg(unsigned long offset)
 {
-    return uniNBaseAddress[offest / 4];
+    return uniNBaseAddress[offset / 4];
 }
 
-void MacRISC2PE::writeUniNReg(unsigned long offest, unsigned long data)
+void MacRISC2PE::writeUniNReg(unsigned long offset, unsigned long data)
 {
-    uniNBaseAddress[offest / 4] = data;
+    uniNBaseAddress[offset / 4] = data;
     eieio();
 }
 
@@ -423,9 +550,12 @@ void MacRISC2PE::getDefaultBusSpeeds(long *numSpeeds, unsigned long **speedList)
     *speedList = macRISC2Speed;
 }
 
-void MacRISC2PE::enableUniNEthernetClock(bool enable)
+void MacRISC2PE::enableUniNEthernetClock(bool enable, IOService *nub)
 {
-    unsigned long regTemp;
+    unsigned long 	regTemp;
+
+	if (enable)
+		configureUniNPCIDevice(nub);		
 
     if (mutex != NULL)
         IOLockLock(mutex);
@@ -447,9 +577,12 @@ void MacRISC2PE::enableUniNEthernetClock(bool enable)
         IOLockUnlock(mutex);
 }
 
-void MacRISC2PE::enableUniNFireWireClock(bool enable)
+void MacRISC2PE::enableUniNFireWireClock(bool enable, IOService *nub)
 {
-    unsigned long regTemp;
+    unsigned long 	regTemp;
+
+	if (enable)
+		configureUniNPCIDevice(nub);		
 
     //IOLog("FWClock, enable = %d kFW = %x\n", enable, kUniNFirewireClockEnable);
   
@@ -492,6 +625,34 @@ void MacRISC2PE::enableUniNFireWireCablePower(bool enable)
     }
 }
 
+void MacRISC2PE::configureUniNPCIDevice (IOService *nub)
+{
+	OSData			*cacheData, *latencyData;
+	IOPCIDevice		*provider;
+
+	if (nub) provider = OSDynamicCast(IOPCIDevice, nub);
+	else provider = NULL;
+	
+	if (provider) {
+		cacheData = (OSData *)provider->getProperty (kIOPCICacheLineSize);
+		latencyData = (OSData *)provider->getProperty (kIOPCITimerLatency);
+	
+		if (cacheData || latencyData) {
+			UInt32				configData;	
+
+			configData = provider->configRead32 (kIOPCIConfigCacheLineSize);
+				
+			if (cacheData) 
+				configData = (configData & 0xFFFFFF00) | *(char *) cacheData->getBytesNoCopy();
+			
+			if (latencyData) 
+				configData = (configData & 0xFFFF00FF) | ((*(char *) latencyData->getBytesNoCopy()) << 8);
+							
+			provider->configWrite32 (kIOPCIConfigCacheLineSize, configData);
+		}
+	}
+	return;
+}
 
 enum
 {
@@ -762,3 +923,403 @@ void MacRISC2PE::PMRegisterDevice(IOService * theNub, IOService * theDevice)
 	}
     }
 }
+
+// -- doPlatformFunction support
+void MacRISC2PE::publishStrings(OSCollection *strings)
+{
+	OSCollectionIterator	*strIter;
+	OSSymbol				*key;
+
+	if (!strings) return;
+
+	strIter = OSCollectionIterator::withCollection(strings);
+
+	if (strIter){
+		while ((key = OSDynamicCast(OSSymbol, strIter->getNextObject())) != 0)
+			publishResource(key, this);
+
+		strIter->release();
+	}
+}
+
+void MacRISC2PE::releaseResources(void)
+{
+	UInt32 i;
+
+	if (fOnDemand) {
+		fOnDemand->release();
+		fOnDemand = 0;
+	}
+
+	for (i=0; i<kListNumLists; i++) {
+		if (fFuncList[i]) {
+			fFuncList[i]->release();
+			fFuncList[i] = 0;
+		}
+	}
+
+	return;
+}
+
+SInt32 MacRISC2PE::parseProvidedFunction(OSString *key, OSData *value)
+{
+	SInt32			numFuncs;
+	UInt32			wordsLeft, phandle, flags, cmdCount, cmdLen, maskLen, valueLen;
+	UInt32			*quadlet;
+	OSData			*cmd;
+	const OSSymbol	*aKey;
+	char			funcNameWithPrefix[160], funcName[128];
+    const char		*myFunc;
+	bool			processingCmdList;
+	
+	processingCmdList = false;
+
+	numFuncs = 0;
+    cmd = 0;
+	cmdCount = 0;
+    	
+	wordsLeft = value->getLength() / sizeof(UInt32);
+
+	quadlet = (UInt32 *)value->getBytesNoCopy();
+	if (quadlet == 0) return(-1);
+	
+	while (wordsLeft >= 3) { // need at least (phandle,flags,cmd) -> 3 words
+		if (processingCmdList)
+			// if cmdCount is zero, done processing list
+			processingCmdList = (cmdCount != 0);
+		else {
+			numFuncs++;
+			phandle = *quadlet++;
+			flags = *quadlet++;
+			wordsLeft -= 2;
+		}
+
+		// If no flags are set this is a useless command
+		if (!flags) return(-1);
+		
+		// Examine the command - not all commands are supported
+		switch (*quadlet) {
+			case kCommandCommandList:
+				DLOG ("parseProvidedFunction got command kCommandCommandList,");
+				processingCmdList = true;
+				cmd = OSData::withBytes ((void *)&phandle, sizeof(UInt32));
+				cmdLen = 1 + kCommandCommandListLength;
+				cmd->appendBytes((void *)quadlet, cmdLen * sizeof(UInt32));
+				quadlet++;
+				cmdCount = *quadlet++;
+				wordsLeft -= cmdLen;
+				DLOG (" cmdCount = %ld\n", cmdCount);
+				break;
+			
+			case kCommandWriteReg32:
+				DLOG ("parseProvidedFunction got command kCommandWriteReg32\n");
+				if (wordsLeft < kCommandWriteReg32Length) return(-1);
+
+				if (processingCmdList)
+					cmdCount--;
+				else
+					cmd = OSData::withBytes ((void *)&phandle, sizeof(UInt32));
+				cmdLen = 1 + kCommandWriteReg32Length;
+				cmd->appendBytes((void *)quadlet, cmdLen * sizeof(UInt32));
+				quadlet += cmdLen;
+				wordsLeft -= cmdLen;
+				break;
+
+			case kCommandConfigRead:
+				DLOG ("parseProvidedFunction got command kCommandConfigRead\n");
+				if (wordsLeft < kCommandConfigReadLength) return(-1);
+
+				if (processingCmdList)
+					cmdCount--;
+				else
+					cmd = OSData::withBytes ((void *)&phandle, sizeof(UInt32));
+				cmdLen = 1 + kCommandConfigReadLength;
+				cmd->appendBytes((void *)quadlet, cmdLen * sizeof(UInt32));
+				quadlet += cmdLen;
+				wordsLeft -= cmdLen;
+				break;
+
+			case kCommandConfigRMW:
+				DLOG ("parseProvidedFunction got command kCommandConfigRMW\n");
+				if (wordsLeft < kCommandConfigRMWLength) return(-1);
+
+				if (processingCmdList)
+					cmdCount--;
+				else
+					cmd = OSData::withBytes ((void *)&phandle, sizeof(UInt32));
+				
+				// Calculate correct length of command, taking into account mask and data length
+				maskLen = quadlet[2];	// get mask array length
+				valueLen = quadlet[3];	// get data array length
+				cmdLen = 1 + kCommandConfigRMWLength + (maskLen / sizeof(UInt32)) + (valueLen / sizeof(UInt32));
+
+				cmd->appendBytes((void *)quadlet, cmdLen * sizeof (UInt32));
+
+				quadlet += cmdLen;
+				wordsLeft -= cmdLen;
+				break;
+
+			default:
+				// unsupported command
+				DLOG("MacRISC2PE got unsupported command\n" \
+				      "\tphandle = %08lx\n" \
+				      "\t  flags = %08lx\n" \
+				      "\tcommand = %08lx\n", phandle, flags, *quadlet);
+				return(-1);
+		}
+
+		if (!cmd) return(-1);
+
+		/*
+		 * If we're not processing a commandList or there aren't enought wordsLeft
+		 * for a full command, then assume we have a complete command and process it.
+		 */
+		if (!processingCmdList || (wordsLeft < 3)) {
+			// Examine flags, add command to appropriate execution list(s)
+			if (flags & kFlagOnInit) ADD_OBJ_TO_SET(cmd, fFuncList[kListOnInit]);
+			if (flags & kFlagOnTerm) ADD_OBJ_TO_SET(cmd, fFuncList[kListOnTerm]);
+	
+			// sleep and wake to be added in later release
+			//if (flags & kFlagOnSleep) ADD_OBJ_TO_SET(cmd, fFuncList[kListOnSleep], -1);
+			//if (flags & kFlagOnWake) ADD_OBJ_TO_SET(cmd, fFuncList[kListOnWake], -1);
+			
+			// On-demand functions need special attention
+			if (flags & kFlagOnDemand) {
+				// Build an OSString to represent this function
+				myFunc = key->getCStringNoCopy() + strlen(kFunctionProvidedPrefix);
+				sprintf(funcName, "%s-%08lx", myFunc, phandle);
+				sprintf(funcNameWithPrefix, "%s%s", kFunctionRequiredPrefix,
+						funcName);
+	
+				if ((aKey = OSSymbol::withCString(funcNameWithPrefix)) == 0)
+					return(-1);
+				
+				DLOG ("parseProvidedFunction registering demand function '%s\n", funcNameWithPrefix);
+				// register for on-demand execution
+				if (fOnDemand) {
+					if (!fOnDemand->setObject(aKey,cmd))
+						return(-1);
+				} else 
+					if ((fOnDemand = OSDictionary::withObjects(
+							&(const OSObject *)cmd,	&aKey, 1, 1)) == 0)
+						return(-1);
+			}
+			cmd->release();
+		}
+	}
+
+	return(numFuncs);
+}
+
+bool MacRISC2PE::performFunctionList(const OSSet *funcList)
+{
+	bool ret;
+	OSCollectionIterator	*iter;
+	const OSData			*aFunc;
+	
+	// Sanity check
+	if (funcList == 0) return(false);
+
+	iter = OSCollectionIterator::withCollection(funcList);
+	if (iter == 0) return(false);
+	
+	ret = true;
+	
+	// Attempt every function in the list.  If any of them fail, return false.
+	while((aFunc = OSDynamicCast(OSData, iter->getNextObject())) != 0)
+		if (!performFunction(aFunc, (void *)1, (void *)0, (void *)0, (void *)0)) ret = false;
+
+	iter->release();
+	return(ret);
+}
+
+bool MacRISC2PE::performFunction(const OSData *func, void *param1 = 0,
+			void *param2 = 0, void *param3 = 0, void *param4 = 0)
+{
+	UInt32		*quadlet, pHandle, offset, data, value, valueLen, mask, maskLen, writeLen, wordsLeft, cmdCount;
+	SInt32		lastCmd;
+	bool		processingCmdList;
+	IOPCIDevice	*nub;
+	
+	if (func == 0) return(false);
+
+	processingCmdList = false;
+
+	cmdCount = 0;
+    	
+	wordsLeft = func->getLength() / sizeof(UInt32);
+	
+	data = 0;
+	nub = NULL;
+	lastCmd = -1;
+	quadlet = (UInt32 *)func->getBytesNoCopy();
+	if (quadlet == 0) return(-1);
+	
+	while (wordsLeft >= 2) { // need at least (phandle,cmd) -> 2 words
+		if (processingCmdList)
+			// if cmdCount is zero, done processing list
+			processingCmdList = (cmdCount != 0);
+		else {
+			pHandle = *quadlet++;
+			wordsLeft--;
+		}
+		
+		// Examine the command - not all commands are supported
+		switch (*quadlet) {
+			case kCommandCommandList:
+				DLOG ("MacRISC2PE::performFunction got command kCommandCommandList,");
+				processingCmdList = true;
+				quadlet++;
+				cmdCount = *quadlet++;
+				wordsLeft -= (1 + kCommandCommandListLength);
+				DLOG (" cmdCount = %ld\n", cmdCount);
+				break;
+			
+			case kCommandWriteReg32:
+				DLOG ("MacRISC2PE::performFunction command kCommandWriteReg32\n");
+				if (processingCmdList)
+					cmdCount--;
+
+				quadlet++;
+				offset = *quadlet++;
+				value = *quadlet++;
+				mask  = *quadlet++;
+				wordsLeft -= (1 + kCommandWriteReg32Length);
+	
+				// If mask isn't all ones, read data and mask it
+				if (mask != 0xFFFFFFFF) {
+					data = readUniNReg (offset);
+					data &= mask;
+					data |= value;
+				} else	// just write the data
+					data = value;
+				
+				DLOG ("MacRISC2PE::performFunction - writing data 0x%lx to Uni-N register offset 0x%lx\n", data, offset);
+				// write the result to the Uni-N register
+				writeUniNReg(offset, data);
+				
+				break;
+	
+			// Currently only handle config reads of 4 bytes or less
+			case kCommandConfigRead:
+				DLOG ("MacRISC2PE::performFunction command kCommandConfigRead\n");
+				if (processingCmdList)
+					cmdCount--;
+
+				quadlet++;
+				offset = *quadlet++;
+				valueLen = *quadlet++;
+				wordsLeft -= (1 + kCommandConfigReadLength);
+				
+				if (valueLen != 4) {
+					IOLog ("MacRISC2PE::performFunction config reads cannot handle anything other than 4 bytes, found length %ld\n", valueLen);
+					return false;
+				}
+	
+				if (!nub) {
+					if (!pHandle) {
+						IOLog ("MacRISC2PE::performFunction config read requires pHandle to locate nub\n");
+						return false;
+					}
+					nub = findNubForPHandle (pHandle);
+					if (!nub) {
+						IOLog ("MacRISC2PE::performFunction config read cannot find nub for pHandle 0x%08lx\n", pHandle);
+						return false;
+					}
+				}
+				
+				// NOTE - code below assumes read of 4 bytes, i.e., valueLen == 4!!
+				data = nub->configRead32 (offset);
+				DLOG ("MacRISC2PE::performFunction config read 0x%lx at offset 0x%lx\n", data, offset);
+				if (param1)
+					*(UInt32 *)param1 = data;
+				
+				lastCmd = kCommandConfigRead;
+				break;
+				
+			// Currently only handle config reads/writes of 4 bytes
+			case kCommandConfigRMW:
+				DLOG ("MacRISC2PE::performFunction command kCommandConfigRMW\n");
+				if (processingCmdList)
+					cmdCount--;
+
+				// data must have been read above
+				if (lastCmd != kCommandConfigRead) {
+					IOLog ("MacRISC2PE::performFunction - config modify/write requires prior read\n");
+					return false;
+				}
+				
+				quadlet++;
+				offset = *quadlet++;
+				maskLen = *quadlet++;
+				valueLen = *quadlet++;
+				writeLen = *quadlet++;
+				wordsLeft -= (1 + kCommandConfigRMWLength + (maskLen / sizeof(UInt32)) + (valueLen / sizeof(UInt32)));
+
+				if (writeLen != 4) {
+					IOLog ("MacRISC2PE::performFunction config read/modify/write cannot handle anything other than 4 bytes, found length %ld\n", writeLen);
+					return false;
+				}
+				
+				// NOTE - code below makes implicit assumption that mask and value are each 4 bytes!!
+				mask = *quadlet++;
+				value = *quadlet++;
+	
+				DLOG("\nMacRISC2PE::performFunction config read/modify/write offset = %08lx maskLen = %08lx dataLen = %08lx writeLen = %08lx, mask = %08lx, data = %08lx\n",
+						offset, maskLen, valueLen, writeLen, mask, value);
+	
+				if (!nub) {
+					if (!pHandle) {
+						IOLog ("MacRISC2PE::performFunction config read/modify/write requires pHandle to locate nub\n");
+						return false;
+					}
+					nub = findNubForPHandle (pHandle);
+					if (!nub) {
+						IOLog ("MacRISC2PE::performFunction config read/modify/write cannot find nub for pHandle 0x%08lx\n", pHandle);
+						return false;
+					}
+				}
+				
+				// data must have been previously read
+				data &= mask;
+				data |= value;
+				
+				DLOG ("MacRISC2PE::performFunction config write data 0x%lx to offset 0x%lx\n", data, offset);
+				nub->configWrite32 (offset, data);
+	
+				break;
+	
+			default:
+				IOLog("MacRISC2PE::performFunction got unsupported command %08lx\n", *quadlet);
+				return(false);
+		}
+	}
+	
+	return(true);
+}
+
+IOPCIDevice* MacRISC2PE::findNubForPHandle( UInt32 pHandleValue )
+{
+	IORegistryIterator*								iterator;
+	IORegistryEntry*								matchingEntry = NULL;
+
+	iterator = IORegistryIterator::iterateOver( gIODTPlane, kIORegistryIterateRecursively );
+
+	if ( iterator == NULL )
+		return( NULL );
+
+	while ( ( matchingEntry = iterator->getNextObject() ) != NULL ) {
+		OSData*						property;
+
+		if ( ( property = OSDynamicCast( OSData, matchingEntry->getProperty( "AAPL,phandle" ) ) ) != NULL )
+			if ( pHandleValue == *( ( UInt32 * ) property->getBytesNoCopy() ) )
+				break;
+	}
+	
+	iterator->release();
+	
+	DLOG ("MacRISC2PE::findNubForPHandle - found matchingEntry '%s'\n", matchingEntry->getName());
+
+	return( OSDynamicCast (IOPCIDevice, matchingEntry));
+}
+
